""" Testing core functionality of pymrio
"""

import os
import sys
from pathlib import Path

import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

import pymrio  # noqa
from pymrio.core.constants import PYMRIO_PATH  # noqa


@pytest.fixture()
def fix_testmrio():
    """Single point to load the test mrio"""

    class TestMRIO:
        testmrio = pymrio.load_test()
        sectors = [
            "food",
            "mining",
            "manufactoring",
            "electricity",
            "construction",
            "trade",
            "transport",
            "other",
        ]
        regions = ["reg1", "reg2", "reg3", "reg4", "reg5", "reg6"]

        Y_cat = [
            "Final consumption expenditure by households",
            "Final consumption expenditure by non-profit "
            "organisations serving households (NPISH)",
            "Final consumption expenditure by government",
            "Gross fixed capital formation",
            "Changes in inventories",
            "Changes in valuables",
            "Export",
        ]

    return TestMRIO


def test_copy(fix_testmrio):
    """Testing the deep copy functionality + naming"""
    tt = fix_testmrio.testmrio
    tt_copy = tt.copy()
    assert tt_copy.name == tt.name + "_copy"
    assert all(tt_copy.Z == tt.Z)
    assert all(tt_copy.emissions.F == tt.emissions.F)
    tt_copy.emissions.F = tt_copy.emissions.F + 2
    tt_copy.Z = tt_copy.Z + 2
    assert all(tt_copy.Z != tt.Z)
    assert all(tt_copy.emissions.F != tt.emissions.F)
    tt_new = tt.copy("new")
    assert tt_new.name == "new"
    e_new = tt.emissions.copy("new")
    assert e_new.name == "new"


def test_get_gross_trade(fix_testmrio):
    tt = fix_testmrio.testmrio
    gross_trade = tt.get_gross_trade()
    flows = gross_trade.bilat_flows
    totals = gross_trade.totals

    reg2mining_exports = (
        tt.Z.loc[("reg2", "mining"), :].sum()
        - tt.Z.loc[("reg2", "mining"), "reg2"].sum()
        + tt.Y.loc[("reg2", "mining"), :].sum()
        - tt.Y.loc[("reg2", "mining"), "reg2"].sum()
    )

    assert reg2mining_exports == pytest.approx(totals.exports.loc["reg2", "mining"])

    for reg in tt.get_regions():
        assert flows.loc[reg, reg].sum().sum() == 0

    reg3trade_imports_from_reg2 = (
        tt.Z.loc[("reg2", "trade"), "reg3"].sum()
        + tt.Y.loc[("reg2", "trade"), "reg3"].sum()
    )
    assert reg3trade_imports_from_reg2 == pytest.approx(
        flows.loc[("reg2", "trade"), "reg3"]
    )


def test_get_index(fix_testmrio):
    """Testing the different options for get_index in core.mriosystem"""
    tt = fix_testmrio.testmrio
    assert all(tt.emissions.F.index == tt.emissions.get_index(as_dict=False))

    F_dict = {ki: ki for ki in tt.emissions.F.index}
    idx_dict = {ki: ki for ki in tt.emissions.get_index(as_dict=True)}
    assert F_dict == idx_dict

    pat1 = {("emis.*", "air"): ("emission alternative", "air")}
    pat1index = tt.emissions.get_index(as_dict=True, grouping_pattern=pat1)
    assert pat1index[tt.emissions.F.index[0]] == list(pat1.values())[0]
    assert pat1index[tt.emissions.F.index[1]] == tt.emissions.F.index[1]

    pat2 = {("emis.*", ".*"): "r"}

    pat2index = tt.emissions.get_index(as_dict=True, grouping_pattern=pat2)
    assert all([True if v == "r" else False for v in pat2index.values()])

    # test one level index

    # pat_single_tpl = {('Value Added',): 'va'}
    pat_single_str = {"Value Added": "va"}
    pat_single_index = tt.factor_inputs.get_index(
        as_dict=True, grouping_pattern=pat_single_str
    )
    assert pat_single_str == pat_single_index


def test_set_index(fix_testmrio):
    tt = fix_testmrio.testmrio
    new_index = ["a", "b"]
    tt.emissions.set_index(new_index)
    assert tt.emissions.get_index()[0] == new_index[0]
    assert tt.emissions.get_index()[1] == new_index[1]


def test_get_sectors(fix_testmrio):
    assert list(fix_testmrio.testmrio.get_sectors()) == fix_testmrio.sectors
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[
            fix_testmrio.sectors.index("construction")
        ]
        == "construction"
    )
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[
            fix_testmrio.sectors.index("food")
        ]
        == "food"
    )
    assert fix_testmrio.testmrio.get_sectors("food") == [
        e if e == "food" else None for e in fix_testmrio.sectors
    ]
    assert (
        fix_testmrio.testmrio.get_sectors(["construction", "food", "a"])[1] == None
    )  # noqa


def test_get_regions(fix_testmrio):
    assert list(fix_testmrio.testmrio.get_regions()) == fix_testmrio.regions

    assert (
        fix_testmrio.testmrio.get_regions(fix_testmrio.regions[:])
        == fix_testmrio.regions
    )

    assert fix_testmrio.testmrio.get_regions("reg4") == [
        e if e == "reg4" else None for e in fix_testmrio.regions
    ]


def test_get_Y_categories(fix_testmrio):
    assert list(fix_testmrio.testmrio.get_Y_categories()) == fix_testmrio.Y_cat

    assert (
        fix_testmrio.testmrio.get_Y_categories(fix_testmrio.Y_cat[:])
        == fix_testmrio.Y_cat
    )

    assert fix_testmrio.testmrio.get_Y_categories("Export") == [
        e if e == "Export" else None for e in fix_testmrio.Y_cat
    ]


def test_rename_regions(fix_testmrio):
    new_reg_name = "New Region 1"
    new_reg_list = ["a1", "a2", "a3", "a4", "a5", "a6"]
    fix_testmrio.testmrio.rename_regions({"reg1": new_reg_name})
    assert fix_testmrio.testmrio.get_regions()[0] == new_reg_name
    fix_testmrio.testmrio.rename_regions(new_reg_list)
    assert fix_testmrio.testmrio.get_regions()[0] == new_reg_list[0]
    assert fix_testmrio.testmrio.get_regions()[2] == new_reg_list[2]


def test_rename_random_sectors(fix_testmrio):
    new_sec_name = "yummy"
    new_sec_list = ["s1", "s2", "s3", "s4", "s5", "s6"]
    fix_testmrio.testmrio.rename_sectors({"food": new_sec_name})
    assert fix_testmrio.testmrio.get_sectors()[0] == new_sec_name
    fix_testmrio.testmrio.rename_sectors(new_sec_list)
    assert fix_testmrio.testmrio.get_sectors()[0] == new_sec_list[0]
    assert fix_testmrio.testmrio.get_sectors()[4] == new_sec_list[4]


def test_rename_sector_with_io_info(fix_testmrio):
    with pytest.raises(ValueError):
        _ = pymrio.get_classification("foo")

    classdata = pymrio.get_classification("test")
    corr_dict = classdata.get_sector_dict(orig="TestMrioName", new="TestMrioCode")
    corr_dict2 = classdata.get_sector_dict(
        orig=fix_testmrio.testmrio.get_sectors(), new="TestMrioCode"
    )

    assert corr_dict == corr_dict2

    fix_testmrio.testmrio.rename_sectors(corr_dict)

    assert (
        fix_testmrio.testmrio.get_sectors()[3]
        == classdata.sectors.iloc[3, :].loc["TestMrioCode"]
    )


def test_rename_Ycat(fix_testmrio):
    new_cat_name = "HouseCons"
    new_cat_list = ["y1", "y2", "y3", "y4", "y5", "y6", "y7"]
    fix_testmrio.testmrio.rename_Y_categories({fix_testmrio.Y_cat[0]: new_cat_name})
    assert fix_testmrio.testmrio.get_Y_categories()[0] == new_cat_name
    fix_testmrio.testmrio.rename_Y_categories(new_cat_list)
    assert fix_testmrio.testmrio.get_Y_categories()[0] == new_cat_list[0]
    assert fix_testmrio.testmrio.get_Y_categories()[-1] == new_cat_list[-1]


def test_copy_and_extensions(fix_testmrio):
    tcp = fix_testmrio.testmrio.copy()
    tcp.remove_extension("Emissions")
    assert len(list(tcp.get_extensions())) == 1
    tcp.remove_extension()
    assert len(list(tcp.get_extensions())) == 0
    assert len(list(fix_testmrio.testmrio.get_extensions())) == 2


def test_get_row_data(fix_testmrio):
    stressor = ("emission_type1", "air")
    tt = fix_testmrio.testmrio.copy().calc_all()
    td = tt.emissions.get_row_data(stressor)["D_exp_reg"]
    md = pd.DataFrame(tt.emissions.D_exp_reg.loc[stressor])
    pdt.assert_frame_equal(td, md)

    for df_name in tt.emissions.get_DataFrame():
        assert df_name in tt.emissions.get_row_data(stressor)


def test_diag_stressor(fix_testmrio):
    stressor_name = ("emission_type1", "air")
    stressor_number = 0
    ext = fix_testmrio.testmrio.emissions
    dext_name = ext.diag_stressor(stressor_name)
    dext_number = ext.diag_stressor(stressor_number)
    pdt.assert_frame_equal(dext_name.F, dext_number.F)

    assert dext_name.F.iloc[0, 0] == ext.F.iloc[stressor_number, 0]
    assert dext_name.F.iloc[1, 1] == ext.F.iloc[stressor_number, 1]
    assert sum(dext_name.F.iloc[0, 1:-1]) == 0
    assert sum(dext_name.F.iloc[1:-1, 0]) == 0


def test_characterize_extension(fix_testmrio):
    factors = pd.read_csv(
        Path(PYMRIO_PATH["test_mrio"] / Path("concordance") / "emissions_charact.tsv"),
        sep="\t",
    )

    shuffled = factors.sample(len(factors.index), random_state=666, axis=0)

    t_uncalc = fix_testmrio.testmrio
    t_calc = fix_testmrio.testmrio.calc_all()
    uncalc_name = "emissions_charact_uncalc"
    ex_uncalc = t_uncalc.emissions.characterize(factors, name=uncalc_name)
    ex_calc = t_uncalc.emissions.characterize(factors)

    assert ex_uncalc.name == uncalc_name
    assert ex_calc.name == t_calc.emissions.name + "_characterized"

    # The test characterization matrix is all in t, the emissions in test are
    # all in kg
    assert ex_calc.unit.loc["total air emissions", "unit"] == "t"
    npt.assert_allclose(
        ex_uncalc.F.loc["total air emissions"].sum(),
        (t_calc.emissions.F.loc[("emission_type1", "air"), :] / 1000).sum(),
    )
    npt.assert_allclose(
        ex_calc.D_imp.loc["total air emissions"].sum(),
        (t_calc.emissions.D_imp.loc[("emission_type1", "air"), :] / 1000).sum(),
    )
    npt.assert_allclose(
        ex_calc.D_exp.loc["air water impact"].sum(),
        (
            (t_calc.emissions.D_exp.loc[("emission_type1", "air"), :] * 2 / 1000)
            + (t_calc.emissions.D_exp.loc[("emission_type2", "water"), :] * 1 / 1000)
        ).sum(),
    )

    # coefficients and multipliers can not characterized directly, so these
    # should be removed and then recalculated

    assert ex_calc.M == None
    assert ex_calc.S == None
    t_calc.impacts = ex_calc
    t_calc.calc_all()
    pdt.assert_series_equal(
        t_calc.impacts.M.loc["total air emissions", :],
        t_calc.emissions.M.loc[("emission_type1", "air"), :] / 1000,
        check_names=False,
    )
    pdt.assert_series_equal(
        t_calc.impacts.S.loc["total air emissions", :],
        t_calc.emissions.S.loc[("emission_type1", "air"), :] / 1000,
        check_names=False,
    )

    with pytest.raises(AssertionError):
        ex_error = t_uncalc.emissions.characterize(
            factors, characterization_factors_column="foo"
        )
    with pytest.raises(AssertionError):
        ex_error = t_uncalc.factor_inputs.characterize(
            factors, characterization_factors_column="foo"
        )

    # testing used characterization matrix
    ret = t_uncalc.emissions.characterize(factors, return_char_matrix=True)
    assert "emissions_type3" not in ret.factors.index

    # testing characterization which do not cover all stressors
    factors_short = factors[
        (factors.stressor == "emission_type1")
        & (factors.impact == "total air emissions")
    ]
    t_calc.short_impacts = t_calc.emissions.characterize(factors_short, name="shorty")
    t_calc.calc_all()

    pdt.assert_series_equal(
        t_calc.short_impacts.S.loc["total air emissions", :],
        t_calc.emissions.S.loc[("emission_type1", "air"), :] / 1000,
        check_names=False,
    )


def test_reset_to_flows(fix_testmrio):
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_to_flows()
    assert tt.A is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_to_flows()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_to_flows(force=True)


def test_reset_all_to_flows(fix_testmrio):
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_all_to_flows()
    assert tt.A is None
    assert tt.emissions.S is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_to_flows()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_to_flows(force=True)


def test_reset_full(fix_testmrio):
    tt = fix_testmrio.testmrio
    assert tt.A is None
    tt.calc_all()
    tt.reset_full()
    assert tt.A is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_full()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_full(force=True)


def test_reset_all_full(fix_testmrio):
    tt = fix_testmrio.testmrio
    assert tt.A is None
    assert tt.emissions.S is None
    tt.calc_all()
    tt.reset_all_full()
    assert tt.A is None
    assert tt.emissions.S is None
    assert tt.emissions.D_exp is None
    tt.Z = None
    with pytest.raises(pymrio.core.mriosystem.ResetError):
        tt.reset_all_full()
    with pytest.warns(pymrio.core.mriosystem.ResetWarning):
        tt.reset_all_full(force=True)


def test_reset_to_coefficients(fix_testmrio):
    tt = fix_testmrio.testmrio
    tt.calc_all()
    tt.reset_all_to_coefficients()
    assert tt.Z is None
    assert tt.emissions.F is None


def test_direct_account_calc(fix_testmrio):
    orig = fix_testmrio.testmrio
    orig.calc_all()

    with pytest.raises(ValueError):
        (D_cba, D_pba, D_imp, D_exp) = pymrio.calc_accounts(
            orig.emissions.S, orig.L, orig.Y
        )

    new = orig.copy().rename_regions({"reg3": "ll", "reg4": "aa"})

    Y_agg = new.Y.groupby(axis=1, level="region", sort=False).agg(sum)

    (D_cba, D_pba, D_imp, D_exp) = pymrio.calc_accounts(new.emissions.S, new.L, Y_agg)

    pdt.assert_frame_equal(orig.emissions.D_cba["reg2"], D_cba["reg2"])
    pdt.assert_frame_equal(orig.emissions.D_exp["reg4"], D_exp["aa"])
    pdt.assert_frame_equal(orig.emissions.D_imp["reg3"], D_imp["ll"])


def test_extension_reset_with_rename(fix_testmrio):
    orig = fix_testmrio.testmrio
    orig.calc_all()

    new = orig.copy()
    new.reset_extensions()
    new.Y = new.Y.loc[:, ["reg2", "reg3"]]
    new.calc_all()

    pdt.assert_frame_equal(new.A, orig.A)

    pdt.assert_frame_equal(new.emissions.D_cba["reg2"], orig.emissions.D_cba["reg2"])
    pdt.assert_frame_equal(new.emissions.D_imp["reg2"], orig.emissions.D_imp["reg2"])

    assert orig.emissions.D_exp_reg.reg1.sum() > new.emissions.D_exp_reg.reg1.sum()

    orig_rename = orig.copy()
    orig_rename.reset_extensions()
    orig_rename.rename_regions({"reg3": "cd", "reg4": "ab"})

    orig_rename.calc_all()
    pdt.assert_frame_equal(
        orig.emissions.D_imp["reg3"], orig_rename.emissions.D_imp["cd"]
    )
    new_rename = orig_rename.copy().reset_extensions()
    new_rename.Y = new_rename.Y.loc[:, ["reg2", "cd", "ab"]]
    new_rename.calc_all()
    pdt.assert_frame_equal(
        new_rename.emissions.D_imp["reg2"], orig_rename.emissions.D_imp["reg2"]
    )
    pdt.assert_frame_equal(
        new_rename.emissions.D_cba["cd"], orig_rename.emissions.D_cba["cd"]
    )
    pdt.assert_frame_equal(
        new_rename.emissions.D_cba["ab"], orig_rename.emissions.D_cba["ab"]
    )
    pdt.assert_frame_equal(
        orig.emissions.D_cba["reg4"], new_rename.emissions.D_cba["ab"]
    )



"""
Pymrio - A python module for automating io calculations and generating reports
==============================================================================

The classes and tools in this module should work with any symetric IO system.

The main class of the module (IOSystem) has attributes .A, .L, ...
corresponding to a standard IO classification.
Data can be assigned directly to the attributes or by calling the
parsing or load functions.

Data storage
------------
Txt files together with a ini file are used for storing data. In addition,
the IOSystem with all data can also be pickled (binary).


Misc
----

Standard abbreviation for that module: mr
Dependencies:

- numpy
- scipy
- pandas
- matplotlib
- docutils (only for IOSystem.report* if format is html and tex - not
            imported otherwise)

:Authors: Konstantin Stadler

:license: BSD 3-Clause License

"""

import logging
import sys

from pymrio.core.fileio import *
from pymrio.core.mriosystem import Extension, IOSystem, concate_extension
from pymrio.tools.ioclass import ClassificationData, get_classification
from pymrio.tools.iodownloader import (
    download_eora26,
    download_exiobase1,
    download_exiobase2,
    download_exiobase3,
    download_gloria,
    download_oecd,
    download_wiod2013,
)
from pymrio.tools.iomath import (
    calc_A,
    calc_accounts,
    calc_e,
    calc_F,
    calc_F_Y,
    calc_gross_trade,
    calc_L,
    calc_M,
    calc_S,
    calc_S_Y,
    calc_x,
    calc_x_from_L,
    calc_Z,
)
from pymrio.tools.iometadata import MRIOMetaData
from pymrio.tools.ioparser import *
from pymrio.tools.ioutil import build_agg_matrix, build_agg_vec
from pymrio.version import __version__



""" test cases for all util functions """

import os
import string
import sys
from collections import namedtuple
from unittest.mock import mock_open, patch

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from faker import Faker

TESTPATH = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(TESTPATH, ".."))

from pymrio.tools.ioutil import build_agg_matrix  # noqa
from pymrio.tools.ioutil import build_agg_vec  # noqa
from pymrio.tools.ioutil import diagonalize_blocks  # noqa
from pymrio.tools.ioutil import find_first_number  # noqa
from pymrio.tools.ioutil import set_block  # noqa
from pymrio.tools.ioutil import sniff_csv_format  # noqa
from pymrio.tools.ioutil import diagonalize_columns_to_sectors, filename_from_url


@pytest.fixture()
def csv_test_files_content():
    test_para = namedtuple("test_para", ["text", "sep", "header_rows", "index_col"])

    class example_csv_content:
        test_contents = [
            test_para(
                text=(
                    """idx1,idx2,col1,col2
                               a,b,1,2
                               c,d,2,4\n"""
                ),
                sep=",",
                header_rows=1,
                index_col=2,
            ),
            test_para(
                text=(
                    """idx1,col1,col2
                               a,1,2
                               c,2,4\n"""
                ),
                sep=",",
                header_rows=1,
                index_col=1,
            ),
            test_para(
                text=(
                    """idx1;col1;col2
                               a;1;2
                               c;2;4\n"""
                ),
                sep=";",
                header_rows=1,
                index_col=1,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;1;2
                               c;2;4\n"""
                ),
                sep=";",
                header_rows=2,
                index_col=1,
            ),
            test_para(
                text=(
                    """|col1|col2
                               |l|d
                               a|f|g
                               c|e|l\n"""
                ),
                sep="|",
                header_rows=None,
                index_col=None,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;f;3
                               c;e;l\n"""
                ),
                sep=";",
                header_rows=None,
                index_col=None,
            ),
            test_para(
                text=(
                    """;col1;col2
                               ;l;d
                               a;f;l
                               c;e;3\n"""
                ),
                sep=";",
                header_rows=3,
                index_col=2,
            ),
        ]

    return example_csv_content


def test_find_first_number():
    """Some tests for finding the first number in a sequence"""
    assert find_first_number([0, 1, 2, 3]) == 0
    assert find_first_number(["a", 1, 2, 3]) == 1
    assert find_first_number(["a", 1, "c", 3]) == 1
    assert find_first_number([None, 1, "c", 3]) == 1
    assert find_first_number(["a", "b"]) == None  # noqa
    assert find_first_number([]) == None  # noqa


def test_dev_read(csv_test_files_content):
    """Tests the csv sniffer"""
    for tests in csv_test_files_content.test_contents:
        with patch(
            "builtins.open", mock_open(read_data=tests.text), create=False
        ) as mo:  # noqa
            res = sniff_csv_format("foo")
        assert res["sep"] == tests.sep
        assert res["nr_index_col"] == tests.index_col
        assert res["nr_header_row"] == tests.header_rows


def test_build_agg_matrix():
    """Tests building the aggreagation matrix"""

    am_num = build_agg_matrix([1, 0, 0, 1, 0])
    expected = np.array([[0.0, 1.0, 1.0, 0.0, 1.0], [1.0, 0.0, 0.0, 1.0, 0.0]])
    am_str = build_agg_matrix(
        ["r2", "r1", "r1", "r2", "r1"], pos_dict={"r1": 0, "r2": 1}
    )

    np.testing.assert_allclose(am_num, expected)
    np.testing.assert_allclose(am_str, expected)


def test_build_agg_vec():
    """Simple test based on the test mrio"""
    vec = build_agg_vec(["OECD", "EU"], path="test", miss="RoW")
    expected = ["OECD", "EU", "OECD", "OECD", "RoW", "RoW"]
    assert vec == expected


def test_diagonalize_blocks():
    """Tests the numpy implementation of diagonalize_blocks"""

    # arr (df):  output (df): (blocks = [x, y, z])
    #     (all letters are index or header)
    #       A B     A A A B B B
    #               x y z x y z
    #     a 3 1     3 0 0 1 0 0
    #     b 4 2     0 4 0 0 2 0
    #     c 5 3     0 0 5 0 0 3
    #     d 6 9     6 0 0 9 0 0
    #     e 7 6     0 7 0 0 6 0
    #     f 8 4     0 0 8 0 0 4

    inp_array = np.array(([3, 1], [4, 2], [5, 3], [6, 9], [7, 6], [8, 4]))
    out_array = np.array(
        (
            [3.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 4.0, 0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 5.0, 0.0, 0.0, 3.0],
            [6.0, 0.0, 0.0, 9.0, 0.0, 0.0],
            [0.0, 7.0, 0.0, 0.0, 6.0, 0.0],
            [0.0, 0.0, 8.0, 0.0, 0.0, 4.0],
        )
    )
    diag_array = diagonalize_blocks(arr=inp_array, blocksize=3)
    np.testing.assert_allclose(diag_array, out_array)


def test_diagonalize_columns_to_sectors():
    inp_array = np.array(([3, 1], [4, 2], [5, 3], [6, 9], [7, 6], [8, 4]))
    out_array = np.array(
        (
            [3.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 4.0, 0.0, 0.0, 2.0, 0.0],
            [0.0, 0.0, 5.0, 0.0, 0.0, 3.0],
            [6.0, 0.0, 0.0, 9.0, 0.0, 0.0],
            [0.0, 7.0, 0.0, 0.0, 6.0, 0.0],
            [0.0, 0.0, 8.0, 0.0, 0.0, 4.0],
        )
    )

    regions = ["A", "B"]
    sectors = ["x", "y", "z"]
    reg_sec_index = pd.MultiIndex.from_product(
        [regions, sectors], names=["region", "sector"]
    )

    inp_df = pd.DataFrame(data=inp_array, index=reg_sec_index, columns=regions)
    inp_df.columns.names = ["region"]

    out_df = pd.DataFrame(data=out_array, index=inp_df.index, columns=reg_sec_index)

    diag_df = diagonalize_columns_to_sectors(inp_df)
    pdt.assert_frame_equal(diag_df, out_df)


def test_set_block():
    """Set block util function"""
    full_arr = np.random.random((10, 10))
    block_arr = np.zeros((2, 2))

    mod_arr = set_block(full_arr, block_arr)
    npt.assert_array_equal(mod_arr[0:2, 0:2], block_arr)
    npt.assert_array_equal(mod_arr[2:4, 2:4], block_arr)
    npt.assert_array_equal(mod_arr[0:2, 3:-1], full_arr[0:2, 3:-1])

    with pytest.raises(ValueError):
        block_arr = np.zeros((3, 3))
        mod_arr = set_block(full_arr, block_arr)

    with pytest.raises(ValueError):
        full_arr = np.random.random((10, 12))
        block_arr = np.zeros((2, 2))
        mod_arr = set_block(full_arr, block_arr)


def test_filename_from_url():
    letters = np.array(list(string.printable))
    fake = Faker()
    for _ in range(100):
        filename = fake.file_name()
        url = fake.url() + filename
        if np.random.rand() > 0.5:
            url += np.random.choice(np.array(["?", "&"])) + "".join(
                np.random.choice(letters, size=np.random.randint(1))
            )

    assert filename_from_url(url) == filename

    test_urls = {
        "Democrat.odp": "https://www.arnold-mann.net/Democrat.odp?refvlmw",
        "heavy.numbers": "http://www.maldonado-mccullough.net/heavy.numbers&dbfgbds",
        "prove.js": "frazier.com/prove.js&vsdfvwevr",
        "conference.pdf": "http://www.dennis.com/conference.pdf&34gegrt4",
    }

    for filename in test_urls.keys():
        assert filename_from_url(test_urls[filename]) == filename



from setuptools import find_packages, setup

exec(open("pymrio/version.py").read())

setup(
    name="pymrio",
    description=(
        "A python module for automating input output "
        "calculations and generating reports"
    ),
    long_description=open("README.rst").read(),
    url="https://github.com/IndEcol/pymrio",
    author="Konstantin Stadler",
    author_email="konstantin.stadler@ntnu.no",
    version=__version__,  # noqa
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7.0",
    package_data={
        "": ["*.txt", "*.dat", "*.doc", "*.rst", "*.json", "*.tsv"],
    },
    # This some needs to be here and in environment.yml (for conda)
    install_requires=[
        "pandas >= 1.5",
        "pyarrow >= 11.0",
        "numpy >= 1.20",
        "matplotlib >= 2.0.0",
        "requests >= 2.18",
        "xlrd > 1.1.0 ",
        "openpyxl >= 3.0.6, < 3.1.1",
        # openpyxl 3.1.1 has a bug that breaks the tests, issue is close, should be fine next release
        "docutils >= 0.14",
        "faker >= 18.4.0",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Utilities",
    ],
)



"""
Methods to load previously save mrios

"""

import collections
import configparser
import json
import logging
import os
import re
import sys
import zipfile
from pathlib import Path

import pandas as pd

from pymrio.core.constants import (
    DEFAULT_FILE_NAMES,
    GENERIC_NAMES,
    PYMRIO_PATH,
    STORAGE_FORMAT,
)
from pymrio.core.mriosystem import Extension, IOSystem
from pymrio.tools.iometadata import MRIOMetaData
from pymrio.tools.ioutil import get_file_para


# Exceptions
class ReadError(Exception):
    """Base class for errors occuring while reading MRIO data"""

    pass


def _get_file_format(file_name):
    """Helper function to get the format of a stored file"""
    # TODO: move to pathlib
    given_extension = Path(file_name).suffix.lstrip(".")
    for format_key, format_extension in STORAGE_FORMAT.items():
        if given_extension.lower() in format_extension:
            table_format = format_key
            break
    else:
        raise ValueError("Unkown format of stored files")
    return format_key


def load_all(path, include_core=True, subfolders=None, path_in_arc=None):
    """Loads a full IO system with all extension in path

    Parameters
    ----------
    path : pathlib.Path or string
        Path or path with para file name for the data to load.
        This must either point to the directory containing the uncompressed
        data or the location of a compressed zip file with the data. In the
        later case and if there are several mrio's in the zip file the
        parameter 'path_in_arc' need to be specifiec to further indicate the
        location of the data in the compressed file.

    include_core : boolean, optional
        If False the load method does not include A, L and Z matrix. This
        significantly reduces the required memory if the purpose is only
        to analyse the results calculated beforehand.

    subfolders: list of pathlib.Path or string, optional
        By default (subfolders=None), all subfolders in path containing a json
        parameter file (as defined in DEFAULT_FILE_NAMES['filepara']:
        metadata.json) are parsed. If only a subset should be used, pass a list
        of names of subfolders. These can either be strings specifying direct
        subfolders of path, or absolute/relative path if the extensions are
        stored at a different location. Both modes can be mixed. If the data
        is read from a zip archive the path must be given as described below in
        'path_in_arc', relative to the root defined in the paramter
        'path_in_arc'. Extensions in a different zip archive must be read
        separately by calling the function 'load' for this extension.

    path_in_arc: string, optional
        Path to the data in the zip file (where the fileparameters file is
        located). path_in_arc must be given without leading dot and slash;
        thus to point to the data in the root of the compressed file pass '',
        for data in e.g. the folder 'emissions' pass 'emissions/'.  Only used
        if parameter 'path' points to an compressed zip file.
        Can be None (default) if there is only one mrio database in the
        zip archive (thus only one file_parameter file as the systemtype entry
        'IOSystem'.

    """

    def clean(varStr):
        """get valid python name from folder"""
        return re.sub(r"\W|^(?=\d)", "_", str(varStr))

    path = Path(path)

    if zipfile.is_zipfile(str(path)):
        with zipfile.ZipFile(file=str(path), mode="r") as zz:
            zipcontent = zz.namelist()
        if path_in_arc:
            path_in_arc = str(path_in_arc)
            if path_in_arc not in zipcontent:
                # Not using os.path.join here b/c this adds the wrong
                # separator when reading the zip in windows
                if path_in_arc != "":
                    path_in_arc = (
                        path_in_arc + "/" + DEFAULT_FILE_NAMES["filepara"]
                    ).replace("//", "/")
                else:
                    path_in_arc = DEFAULT_FILE_NAMES["filepara"]

                if path_in_arc not in zipcontent:
                    raise ReadError(
                        "File parameter file {} not found in {}. "
                        "Tip: specify fileparameter filename "
                        'through "path_in_arc" if different '
                        "from default.".format(DEFAULT_FILE_NAMES["filepara"], path)
                    )

        else:
            with zipfile.ZipFile(file=str(path), mode="r") as zz:
                fpfiles = [
                    f
                    for f in zz.namelist()
                    if os.path.basename(f) == DEFAULT_FILE_NAMES["filepara"]
                    and json.loads(zz.read(f).decode("utf-8"))["systemtype"]
                    == "IOSystem"
                ]
            if len(fpfiles) == 0:
                raise ReadError(
                    "File parameter file {} not found in {}. "
                    "Tip: specify fileparameter filename "
                    'through "path_in_arc" if different '
                    "from default.".format(DEFAULT_FILE_NAMES["filepara"], path)
                )
            elif len(fpfiles) > 1:
                raise ReadError(
                    "Mulitple mrio archives found in {}. "
                    "Specify one by the "
                    'parameter "path_in_arc"'.format(path)
                )
            else:
                path_in_arc = os.path.dirname(fpfiles[0])

        logging.debug(
            "Expect file parameter-file at {} in {}".format(path_in_arc, path)
        )

    io = load(path, include_core=include_core, path_in_arc=path_in_arc)

    if zipfile.is_zipfile(str(path)):
        root_in_zip = os.path.dirname(path_in_arc)
        if subfolders is None:
            subfolders = {
                Path(os.path.relpath(os.path.dirname(p), root_in_zip)).as_posix()
                for p in zipcontent
                if p.startswith(root_in_zip) and os.path.dirname(p) != root_in_zip
            }

        for subfolder_name in subfolders:
            if subfolder_name not in zipcontent + list(
                {os.path.dirname(p) for p in zipcontent}
            ):
                # Not using os.path.join here b/c this adds the wrong
                # separator when reading the zip in windows
                subfolder_full = root_in_zip + "/" + subfolder_name
                subfolder_full = subfolder_full.replace("//", "/")
            else:
                subfolder_full = subfolder_name
            subfolder_name = os.path.basename(os.path.normpath(subfolder_name))

            if subfolder_name not in zipcontent:
                subfolder_full_meta = (
                    subfolder_full + "/" + DEFAULT_FILE_NAMES["filepara"]
                )
            else:
                subfolder_full_meta = subfolder_full

            if subfolder_full_meta in zipcontent:
                ext = load(
                    path, include_core=include_core, path_in_arc=subfolder_full_meta
                )
                setattr(io, clean(subfolder_name), ext)
                io.meta._add_fileio(
                    "Added satellite account " "from {}".format(subfolder_full)
                )
            else:
                continue

    else:
        if subfolders is None:
            subfolders = [d for d in path.iterdir() if d.is_dir()]

        for subfolder_name in subfolders:
            if not os.path.exists(str(subfolder_name)):
                subfolder_full = path / subfolder_name
            else:
                subfolder_full = subfolder_name
            subfolder_name = os.path.basename(os.path.normpath(subfolder_name))

            if not os.path.isfile(str(subfolder_full)):
                subfolder_full_meta = subfolder_full / DEFAULT_FILE_NAMES["filepara"]
            else:
                subfolder_full_meta = subfolder_full

            if subfolder_full_meta.exists():
                ext = load(subfolder_full, include_core=include_core)
                setattr(io, clean(subfolder_name), ext)
                io.meta._add_fileio(
                    "Added satellite account " "from {}".format(subfolder_full)
                )
            else:
                continue

    return io


def load(path, include_core=True, path_in_arc=""):
    """Loads a IOSystem or Extension previously saved with pymrio

    This function can be used to load a IOSystem or Extension specified in a
    metadata file (as defined in DEFAULT_FILE_NAMES['filepara']: metadata.json)

    The format of DataFrames (tables) are determined by the file extension.
    Avaialble formats are defined in constant.py - STORAGE_FORMAT.
    If using the save/save_all functionality from the mriosystem it just works.

    Parameters
    ----------
    path : pathlib.Path or string
        Path or path with para file name for the data to load. This must
        either point to the directory containing the uncompressed data or
        the location of a compressed zip file with the data. In the
        later case the parameter 'path_in_arc' need to be specific to
        further indicate the location of the data in the compressed file.

    include_core : boolean, optional
        If False the load method does not include A, L and Z matrix. This
        significantly reduces the required memory if the purpose is only
        to analyse the results calculated beforehand.

    path_in_arc: string, optional
        Path to the data in the zip file (where the fileparameters file is
        located). path_in_arc must be given without leading dot and slash;
        thus to point to the data in the root of the compressed file pass '',
        for data in e.g. the folder 'emissions' pass 'emissions/'.  Only used
        if parameter 'path' points to an compressed zip file.

    Returns
    -------

        IOSystem or Extension class depending on systemtype in the json file
        None in case of errors

    """
    path = Path(path)

    if not path.exists():
        raise ReadError("Given path does not exist")

    file_para = get_file_para(path=path, path_in_arc=path_in_arc)

    if file_para.content["systemtype"] == GENERIC_NAMES["iosys"]:
        if zipfile.is_zipfile(str(path)):
            # Not using os.path.join here b/c this adds the wrong
            # separator when reading the zip in windows
            if file_para.folder != "":
                metadata_folder = (
                    file_para.folder + "/" + DEFAULT_FILE_NAMES["metadata"]
                ).replace("//", "/")
            else:
                metadata_folder = DEFAULT_FILE_NAMES["metadata"]

            ret_system = IOSystem(
                meta=MRIOMetaData(location=path, path_in_arc=metadata_folder)
            )
            ret_system.meta._add_fileio(
                "Loaded IO system from {} - {}".format(path, path_in_arc)
            )
        else:
            ret_system = IOSystem(
                meta=MRIOMetaData(location=path / DEFAULT_FILE_NAMES["metadata"])
            )
            ret_system.meta._add_fileio("Loaded IO system from {}".format(path))

    elif file_para.content["systemtype"] == GENERIC_NAMES["ext"]:
        ret_system = Extension(file_para.content["name"])

    else:
        raise ReadError("Type of system no defined in the file parameters")
        return None

    for key in file_para.content["files"]:
        if not include_core and key not in ["A", "L", "Z"]:
            continue

        file_name = file_para.content["files"][key]["name"]
        nr_index_col = file_para.content["files"][key]["nr_index_col"]
        nr_header = file_para.content["files"][key]["nr_header"]
        _index_col = list(range(int(nr_index_col)))
        _header = list(range(int(nr_header)))
        _index_col = 0 if _index_col == [0] else _index_col
        _header = 0 if _header == [0] else _header

        if key == "FY":  # Legacy code to read data saved with version < 0.4
            key = "F_Y"

        if zipfile.is_zipfile(str(path)):
            # Not using os.path.join here b/c this adds the wrong
            # separator when reading the zip in windows
            if file_para.folder != "":
                full_file_name = file_para.folder + "/" + file_name
                full_file_name = full_file_name.replace("//", "/")
            else:
                full_file_name = file_name
            logging.info("Load data from {}".format(full_file_name))

            with zipfile.ZipFile(file=str(path)) as zf:
                file_format = _get_file_format(full_file_name)
                if file_format == "txt":
                    setattr(
                        ret_system,
                        key,
                        pd.read_csv(
                            zf.open(full_file_name),
                            index_col=_index_col,
                            header=_header,
                            sep="\t",
                        ),
                    )
                elif file_format == "pickle":
                    setattr(ret_system, key, pd.read_pickle(zf.open(full_file_name)))
                elif file_format == "parquet":
                    setattr(ret_system, key, pd.read_parquet(zf.open(full_file_name)))
                else:
                    ValueError("Unknown file format")

        else:
            full_file_name = path / file_name
            logging.info("Load data from {}".format(full_file_name))
            file_format = _get_file_format(full_file_name)
            if file_format == "txt":
                setattr(
                    ret_system,
                    key,
                    pd.read_csv(
                        full_file_name,
                        index_col=_index_col,
                        header=_header,
                        sep="\t",
                    ),
                )
            elif file_format == "pickle":
                setattr(ret_system, key, pd.read_pickle(full_file_name))
            elif file_format == "parquet":
                setattr(ret_system, key, pd.read_parquet(full_file_name))
            else:
                ValueError("Unknown file format")
    return ret_system


def archive(
    source,
    archive,
    path_in_arc=None,
    remove_source=False,
    compression=zipfile.ZIP_DEFLATED,
    compresslevel=-1,
):
    """Archives a MRIO database as zip file

    This function is a wrapper around zipfile.write,
    to ease the writing of an archive and removing the source data.

    Note
    ----
    In contrast to zipfile.write, this function raises an
    error if the data (path + filename) are identical in the zip archive.
    Background: the zip standard allows that files with the same name and path
    are stored side by side in a zip file. This becomes an issue when unpacking
    this files as they overwrite each other upon extraction.

    Parameters
    ----------

    source: str or pathlib.Path or list of these
        Location of the mrio data (folder).
        If not all data should be archived, pass a list of
        all files which should be included in the archive (absolute path)

    archive: str or pathlib.Path
        Full path with filename for the archive.

    path_in_arc: string, optional
        Path within the archive zip file where data should be stored.
        'path_in_arc' must be given without leading dot and slash.
        Thus to point to the data in the root of the compressed file pass '',
        for data in e.g. the folder 'mrio_v1' pass 'mrio_v1/'.
        If None (default) data will be stored in the root of the archive.

    remove_source: boolean, optional
        If True, deletes the source file from the disk (all files
        specified in 'source' or the specified directory, depending if a
        list of files or directory was passed). If False, leaves the
        original files on disk. Also removes all empty directories
        in source including source.

    compression: ZIP compression method, optional
        This is passed to zipfile.write. By default it is set to ZIP_DEFLATED.
        NB: This is different from the zipfile default (ZIP_STORED) which would
        not give any compression. See
        https://docs.python.org/3/library/zipfile.html#zipfile-objects for
        further information. Depending on the value given here additional
        modules might be necessary (e.g. zlib for ZIP_DEFLATED). Futher
        information on this can also be found in the zipfile python docs.

    compresslevel: int, optional
        This is passed to zipfile.write and specifies the compression level.
        Acceptable values depend on the method specified at the parameter
        'compression'.  By default, it is set to -1 which gives a compromise
        between speed and size for the ZIP_DEFLATED compression (this is
        internally interpreted as 6 as described here:
        https://docs.python.org/3/library/zlib.html#zlib.compressobj )
        NB: This is only used if python version >= 3.7

    Raises
    ------
    FileExistsError: In case a file to be archived already present in the
    archive.

    """
    archive = Path(archive)

    if type(source) is not list:
        source_root = str(source)
        source_files = [f for f in Path(source).glob("**/*") if f.is_file()]
    else:
        source_root = os.path.commonpath([str(f) for f in source])
        source_files = [Path(f) for f in source]

    path_in_arc = "" if not path_in_arc else path_in_arc

    # Not using os.path.join here b/c this adds the wrong
    # separator when reading the zip in windows
    arc_file_names = {
        str(f): (path_in_arc + "/" + str(f.relative_to(source_root))).replace("//", "/")
        for f in source_files
    }

    if archive.exists():
        with zipfile.ZipFile(file=str(archive), mode="r") as zf:
            already_present = zf.namelist()
        duplicates = {
            ff: zf for ff, zf in arc_file_names.items() if zf in already_present
        }

        if duplicates:
            raise FileExistsError(
                "These files already exists in {arc} for "
                'path_in_arc "{pa}":\n  {filelist}'.format(
                    pa=path_in_arc,
                    arc=archive,
                    filelist="\n  ".join(duplicates.values()),
                )
            )

    if sys.version_info.major == 3 and sys.version_info.minor >= 7:
        zip_open_para = dict(
            file=str(archive),
            mode="a",
            compression=compression,
            compresslevel=compresslevel,
        )
    else:
        zip_open_para = dict(file=str(archive), mode="a", compression=compression)

    with zipfile.ZipFile(**zip_open_para) as zz:
        for fullpath, zippath in arc_file_names.items():
            zz.write(str(fullpath), str(zippath))

    if remove_source:
        for f in source_files:
            os.remove(str(f))

        for root, dirs, files in os.walk(source_root, topdown=False):
            for name in dirs:
                dir_path = os.path.join(root, name)
                if not os.listdir(dir_path):
                    os.rmdir(os.path.join(root, name))
        try:
            os.rmdir(source_root)
        except OSError:
            pass


def _load_all_ini_based_io(path, **kwargs):  # pragma: no cover
    """DEPRECATED: For convert a previous version to the new json format

    Loads the whole IOSystem with Extensions given in path

    This just calls pymrio.load with recursive = True. Apart from that the
    same parameter can as for .load can be used.

    Parameters
    ----------
    path : string
        path or ini file name for the data to load
    **kwargs : key word arguments, optional
            This will be passed directly to the load method

    Returns
    -------
    IOSystem
    None in case of errors
    """
    return _load_ini_based_io(path, recursive=True, **kwargs)


def _load_ini_based_io(
    path,
    recursive=False,
    ini=None,
    subini={},
    include_core=True,
    only_coefficients=False,
):  # pragma: no cover
    """DEPRECATED: For convert a previous version to the new json format

    Loads a IOSystem or Extension from a ini files

    This function can be used to load a IOSystem or Extension specified in a
    ini file. DataFrames (tables) are loaded from text or binary pickle files.
    For the latter, the extension .pkl or .pickle is assumed, in all other case
    the tables are assumed to be in .txt format.

    Parameters
    ----------

    path : string
        path or ini file name for the data to load

    recursive : boolean, optional
        If True, load also the data in the subfolders and add them as
        extensions to the IOSystem (in that case path must point to the root).
        Only first order subfolders are considered (no subfolders in
        subfolders) and if a folder does not contain a ini file it's skipped.
        Use the subini parameter in case of multiple ini files in a subfolder.
        Attribute name of the extension in the IOSystem are based on the
        subfolder name.  Default is False

    ini : string, optional
        If there are several ini files in the root folder, take this one for
        loading the data If None (default) take the ini found in the folder,
        error if several are found

    subini : dict, optional
        If there are multiple ini in the subfolder, use the ini given in the
        dict.  Format: 'subfoldername':'ininame' If a key for a subfolder is
        not found or None (default), the ini found in the folder will be taken,
        error if several are found

    include_core : boolean, optional
        If False the load method does not include A, L and Z matrix. This
        significantly reduces the required memory if the purpose is only
        to analyse the results calculated beforehand.

    Returns
    -------

        IOSystem or Extension class depending on systemtype in the ini file
        None in case of errors

    """

    # check path and given parameter
    ini_file_name = None

    path = os.path.abspath(os.path.normpath(path))

    if os.path.splitext(path)[1] == ".ini":
        (path, ini_file_name) = os.path.split(path)

    if ini:
        ini_file_name = ini

    if not os.path.exists(path):
        raise ReadError("Given path does not exist")
        return None

    if not ini_file_name:
        _inifound = False
        for file in os.listdir(path):
            if os.path.splitext(file)[1] == ".ini":
                if _inifound:
                    raise ReadError("Found multiple ini files in folder - specify one")
                    return None
                ini_file_name = file
                _inifound = True

    # read the ini
    io_ini = configparser.RawConfigParser()
    io_ini.optionxform = lambda option: option

    io_ini.read(os.path.join(path, ini_file_name))

    systemtype = io_ini.get("systemtype", "systemtype", fallback=None)
    name = io_ini.get("meta", "name", fallback=os.path.splitext(ini_file_name)[0])

    if systemtype == "IOSystem":
        ret_system = IOSystem(name=name)
    elif systemtype == "Extension":
        ret_system = Extension(name=name)
    else:
        raise ReadError("System not defined in ini")
        return None

    for key in io_ini["meta"]:
        setattr(ret_system, key, io_ini.get("meta", key, fallback=None))

    for key in io_ini["files"]:
        if "_nr_index_col" in key:
            continue
        if "_nr_header" in key:
            continue

        if not include_core:
            not_to_load = ["A", "L", "Z"]
            if key in not_to_load:
                continue

        if only_coefficients:
            _io = IOSystem()
            if key not in _io.__coefficients__ + ["unit"]:
                continue

        file_name = io_ini.get("files", key)
        nr_index_col = io_ini.get("files", key + "_nr_index_col", fallback=None)
        nr_header = io_ini.get("files", key + "_nr_header", fallback=None)

        if (nr_index_col is None) or (nr_header is None):
            raise ReadError(
                "Index or column specification missing for {}".format(str(file_name))
            )
            return None

        _index_col = list(range(int(nr_index_col)))
        _header = list(range(int(nr_header)))

        if _index_col == [0]:
            _index_col = 0
        if _header == [0]:
            _header = 0
        file = os.path.join(path, file_name)
        logging.info("Load data from {}".format(file))

        if (
            os.path.splitext(file)[1] == ".pkl"
            or os.path.splitext(file)[1] == ".pickle"
        ):
            setattr(ret_system, key, pd.read_pickle(file))
        else:
            setattr(
                ret_system,
                key,
                pd.read_csv(file, index_col=_index_col, header=_header, sep="\t"),
            )

    if recursive:
        # look for subfolder in the given path
        subfolder_list = os.walk(path).__next__()[1]

        # loop all subfolder and append extension based on
        # ini file in subfolder
        for subfolder in subfolder_list:
            subini_file_name = subini.get(subfolder)
            subpath = os.path.abspath(os.path.join(path, subfolder))

            if not subini_file_name:
                _inifound = False
                for file in os.listdir(subpath):
                    if os.path.splitext(file)[1] == ".ini":
                        if _inifound:
                            raise ReadError(
                                "Found multiple ini files in subfolder "
                                "{} - specify one".format(subpath)
                            )
                            return None
                        subini_file_name = file
                        _inifound = True
            if not _inifound:
                continue

            # read the ini
            subio_ini = configparser.RawConfigParser()
            subio_ini.optionxform = lambda option: option

            subio_ini.read(os.path.join(subpath, subini_file_name))

            systemtype = subio_ini.get("systemtype", "systemtype", fallback=None)
            name = subio_ini.get(
                "meta", "name", fallback=os.path.splitext(subini_file_name)[0]
            )

            if systemtype == "IOSystem":
                raise ReadError(
                    "IOSystem found in subfolder {} - "
                    "only extensions expected".format(subpath)
                )
                return None
            elif systemtype == "Extension":
                sub_system = Extension(name=name)
            else:
                raise ReadError("System not defined in ini")
                return None

            for key in subio_ini["meta"]:
                setattr(sub_system, key, subio_ini.get("meta", key, fallback=None))

            for key in subio_ini["files"]:
                if "_nr_index_col" in key:
                    continue
                if "_nr_header" in key:
                    continue

                if only_coefficients:
                    _ext = Extension("temp")
                    if key not in _ext.__coefficients__ + ["unit"]:
                        continue

                file_name = subio_ini.get("files", key)
                nr_index_col = subio_ini.get(
                    "files", key + "_nr_index_col", fallback=None
                )
                nr_header = subio_ini.get("files", key + "_nr_header", fallback=None)

                if (nr_index_col is None) or (nr_header is None):
                    raise ReadError(
                        "Index or column specification missing "
                        "for {}".format(str(file_name))
                    )
                    return None

                _index_col = list(range(int(nr_index_col)))
                _header = list(range(int(nr_header)))

                if _index_col == [0]:
                    _index_col = 0
                if _header == [0]:
                    _header = 0
                file = os.path.join(subpath, file_name)
                logging.info("Load data from {}".format(file))
                if (
                    os.path.splitext(file)[1] == ".pkl"
                    or os.path.splitext(file)[1] == ".pickle"
                ):
                    setattr(sub_system, key, pd.read_pickle(file))
                else:
                    setattr(
                        sub_system,
                        key,
                        pd.read_csv(
                            file, index_col=_index_col, header=_header, sep="\t"
                        ),
                    )

                # get valid python name from folder
                def clean(varStr):
                    return re.sub(r"\W|^(?=\d)", "_", str(varStr))

                setattr(ret_system, clean(subfolder), sub_system)

    return ret_system


def load_test():
    """Returns a small test MRIO

    The test system contains:

        - six regions,
        - seven sectors,
        - seven final demand categories
        - two extensions (emissions and factor_inputs)

    The test system only contains Z, Y, F, F_Y. The rest can be calculated with
    calc_all()

    Notes
    -----

        For development: This function can be used as an example of
        how to parse an IOSystem

    Returns
    -------

    IOSystem

    """

    # row_header:
    #    number of rows containing header on the top of the file (for the
    #    columns)
    # col_header:
    #    number of cols containing header on the beginning of the file (for the
    #    rows)
    # row and columns header contain also the row for the units, this are
    # afterwards safed as a extra dataframe
    #
    # unit_col: column containing the unit for the table
    file_data = collections.namedtuple(
        "file_data", ["file_name", "row_header", "col_header", "unit_col"]
    )

    # file names and header specs of the system
    test_system = dict(
        Z=file_data(
            file_name="trade_flows_Z.txt", row_header=2, col_header=3, unit_col=2
        ),
        Y=file_data(
            file_name="finald_demand_Y.txt", row_header=2, col_header=3, unit_col=2
        ),
        fac=file_data(
            file_name="factor_input.txt", row_header=2, col_header=2, unit_col=1
        ),
        emissions=file_data(
            file_name="emissions.txt", row_header=2, col_header=3, unit_col=2
        ),
        FDemissions=file_data(
            file_name="FDemissions.txt", row_header=2, col_header=3, unit_col=2
        ),
    )

    meta_rec = MRIOMetaData(location=PYMRIO_PATH["test_mrio_data"])

    # read the data into a dicts as pandas.DataFrame
    data = {
        key: pd.read_csv(
            os.path.join(PYMRIO_PATH["test_mrio_data"], test_system[key].file_name),
            index_col=list(range(test_system[key].col_header)),
            header=list(range(test_system[key].row_header)),
            sep="\t",
        )
        for key in test_system
    }

    meta_rec._add_fileio("Load test_mrio from {}".format(PYMRIO_PATH["test_mrio_data"]))

    # distribute the data into dics which can be passed to
    # the IOSystem. To do so, some preps are necessary:
    # - name the header data
    # - save unit in own dataframe and drop unit from the tables

    trade = dict(Z=data["Z"], Y=data["Y"])
    factor_inputs = dict(F=data["fac"])
    emissions = dict(F=data["emissions"], F_Y=data["FDemissions"])

    trade["Z"].index.names = ["region", "sector", "unit"]
    trade["Z"].columns.names = ["region", "sector"]
    trade["unit"] = pd.DataFrame(trade["Z"].iloc[:, 0].reset_index(level="unit").unit)
    trade["Z"].reset_index(level="unit", drop=True, inplace=True)

    trade["Y"].index.names = ["region", "sector", "unit"]
    trade["Y"].columns.names = ["region", "category"]
    trade["Y"].reset_index(level="unit", drop=True, inplace=True)

    factor_inputs["name"] = "Factor Inputs"
    factor_inputs["F"].index.names = [
        "inputtype",
        "unit",
    ]
    factor_inputs["F"].columns.names = ["region", "sector"]
    factor_inputs["unit"] = pd.DataFrame(
        factor_inputs["F"].iloc[:, 0].reset_index(level="unit").unit
    )
    factor_inputs["F"].reset_index(level="unit", drop=True, inplace=True)

    emissions["name"] = "Emissions"
    emissions["F"].index.names = [
        "stressor",
        "compartment",
        "unit",
    ]
    emissions["F"].columns.names = ["region", "sector"]
    emissions["unit"] = pd.DataFrame(
        emissions["F"].iloc[:, 0].reset_index(level="unit").unit
    )
    emissions["F"].reset_index(level="unit", drop=True, inplace=True)
    emissions["F_Y"].index.names = ["stressor", "compartment", "unit"]
    emissions["F_Y"].columns.names = ["region", "category"]
    emissions["F_Y"].reset_index(level="unit", drop=True, inplace=True)

    # the population data - this is optional (None can be passed if no data is
    # available)
    popdata = pd.read_csv(
        PYMRIO_PATH["test_mrio_data"] / "population.txt",
        index_col=0,
        sep="\t",
    ).astype(float)

    return IOSystem(
        Z=data["Z"],
        Y=data["Y"],
        unit=trade["unit"],
        meta=meta_rec,
        factor_inputs=factor_inputs,
        emissions=emissions,
        population=popdata,
    )
